//   Copyright (c)  2007  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


// Source code for class DistMPolyInlPP

#include "CoCoA/DenseUPolyClean.H"
#include <algorithm>
using std::copy; // copy ctor
#include "CoCoA/error.H"
#include "CoCoA/ZZ.H"  // uses binomial

//#include <vector>
using std::vector;
//#include <memory>
using std::auto_ptr;
//#include <cstddef>
using std::size_t;
#include <algorithm>
using std::max;
//using std::swap;
#include <iostream>
using std::ostream;


namespace CoCoA
{

  DenseUPolyClean::DenseUPolyClean(const ring& R):
      myCoeffRingValue(R)
  {
    myDegPlus1Value = 0;
  }


  DenseUPolyClean::~DenseUPolyClean()
  {
  }


  DenseUPolyClean::DenseUPolyClean(const DenseUPolyClean& copy):
      myCoeffRingValue(copy.myCoeffRingValue),
      myDegPlus1Value(copy.myDegPlus1Value),
      mySizeValue(copy.mySizeValue),
      myCoeffsValue(copy.myCoeffsValue)
  {
    //    copy(rawcopy.begin(), rawcopy.end(), rawlhs.back_inserter());
  }


  DenseUPolyClean& DenseUPolyClean::operator=(const DenseUPolyClean& rhs)
  {
    if (this == &rhs) return *this;
    DenseUPolyClean copy(rhs);
    ourSwap(*this, copy);
    return *this;
  }


  DenseUPolyClean& DenseUPolyClean::operator=(long rhs)
  { // exception safe
    if (rhs == 0)
    {
      myAssignZero();
      return *this; // to avoid needless alloc/free
    }
    myResize(1);
    RingElem x(myCoeffRingValue, rhs);
    swap(myCoeffsValue[0], x);
    myDegPlus1Value = 1;
    return *this;
  }


  DenseUPolyClean& DenseUPolyClean::operator=(const ZZ& rhs)
  { // exception safe
    //    myAssignZero();  // this would probably save memory
    if (rhs == 0)
    {
      myAssignZero();
      return *this; // to avoid needless alloc/free
    }
    RingElem x(myCoeffRingValue, rhs);
    swap(myCoeffsValue[0], x);
    myDegPlus1Value = 1;
    return *this;
  }


  bool IsCompatible(const DenseUPolyClean& f, const DenseUPolyClean& g)
  {
    return (f.myCoeffRingValue == g.myCoeffRingValue);
  }


//----------------------------------------------------------------------

  void DenseUPolyClean::ourSwap(DenseUPolyClean& f, DenseUPolyClean& g)
  {
    CoCoA_ASSERT(IsCompatible(f, g));
    std::swap(f.myCoeffsValue, g.myCoeffsValue);
    std::swap(f.myDegPlus1Value, g.myDegPlus1Value);
    std::swap(f.mySizeValue, g.mySizeValue);
  }


  void DenseUPolyClean::myAssignZero()
  { // NO THROW GUARANTEE
    myDegPlus1Value = 0;
  }


  void DenseUPolyClean::myResize(std::size_t NewSize)
  {
    myCoeffsValue.resize(NewSize, zero(myCoeffRingValue));
    mySizeValue = NewSize;
  }


  void DenseUPolyClean::myResetDeg()
  {
    size_t i = myDegPlus1Value;
    for (; i!=0; --i)
      if (!IsZero(myCoeffsValue[i-1])) break;
    myDegPlus1Value = i;  // either i==0 or myCoeffsValue[i-1]!=0
  }
  

  void DenseUPolyClean::myAssignZeroCoeff(std::size_t d)
  { // exception safe?
    if (d >= myDegPlus1Value)  return;
    // d <= deg
    myCoeffsValue[d] = 0;
    if (d == myDegPlus1Value-1)
      myResetDeg();
  }


  void DenseUPolyClean::myAssignNonZeroCoeff(ConstRefRingElem c, std::size_t d)
  { // exception safe
    CoCoA_ASSERT(mySize()>=d);
    RingElem x(c);
    swap(myCoeffsValue[d], x);
    if (d >= myDegPlus1Value)
    {
      for (size_t i=myDegPlus1Value; i<d; ++i)
        if (!IsZero(myCoeffsValue[i]))
          myCoeffsValue[i] = 0;
      myDegPlus1Value = d+1;
    }
  }


  ConstRefRingElem LC(const DenseUPolyClean& f)
  {
    CoCoA_ASSERT(!IsZero(f));
    CoCoA_ASSERT(!IsZero(f.myCoeffsValue[f.myDegPlus1Value]));
    return f.myCoeffsValue[f.myDegPlus1Value];
  }


  std::size_t DenseUPolyClean::myDegPlus1() const
  {
    //    CoCoA_ASSERT(!IsZero(*this));
    return myDegPlus1Value;
  }
  

  std::size_t DenseUPolyClean::mySize() const
  {
    return mySizeValue;
  }


  void DenseUPolyClean::myNegate()
  {
    for (size_t i=0 ; i < myDegPlus1Value ; ++i)
      myCoeffRingValue->myNegate(raw(myCoeffsValue[i]), raw(myCoeffsValue[i]));   // MIGHT THROW???
  }


  void add(DenseUPolyClean& lhs, const DenseUPolyClean& f, const DenseUPolyClean& g)
  {
    CoCoA_ASSERT(IsCompatible(lhs, g) && IsCompatible(f, g));
    const ring& R = lhs.myCoeffRingValue;
    DenseUPolyClean ans(R); // ans.myDegPlus1Value is 0
    ans.myResize(max(g.myDegPlus1Value, f.myDegPlus1Value)+1);

    size_t i = 0;
    if (f.myDegPlus1Value >= g.myDegPlus1Value)
    {
      for (; i<g.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i] + g.myCoeffsValue[i];
      for (; i<f.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i];
    }
    else
    {
      for (; i<f.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i] + g.myCoeffsValue[i];
      for (; i<g.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = g.myCoeffsValue[i];
    }
    ans.mySizeValue = ans.myCoeffsValue.size();
    ans.myDegPlus1Value = max(f.myDegPlus1Value, g.myDegPlus1Value);
    if (f.myDegPlus1Value == g.myDegPlus1Value)
      ans.myResetDeg();
    swap(lhs, ans);
  }


  void sub(DenseUPolyClean& lhs, const DenseUPolyClean& f, const DenseUPolyClean& g)
  {
    CoCoA_ASSERT(IsCompatible(lhs, g) && IsCompatible(f, g));
    const ring& R = lhs.myCoeffRingValue;
    DenseUPolyClean ans(R); // ans.myDegPlus1Value is 0
    ans.myResize(max(g.myDegPlus1Value, f.myDegPlus1Value)+1);

    size_t i = 0;
    if (f.myDegPlus1Value >= g.myDegPlus1Value)
    {
      for (; i<g.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i] - g.myCoeffsValue[i];
      for (; i<f.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i];
    }
    else
    {
      for (; i<f.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = f.myCoeffsValue[i] - g.myCoeffsValue[i];
      for (; i<g.myDegPlus1Value; ++i)
        ans.myCoeffsValue[i] = - g.myCoeffsValue[i];
    }
    ans.mySizeValue = ans.myCoeffsValue.size();
    ans.myDegPlus1Value = max(f.myDegPlus1Value, g.myDegPlus1Value);
    if (f.myDegPlus1Value == g.myDegPlus1Value)
      ans.myResetDeg();
    swap(lhs, ans);
  }


  bool div(DenseUPolyClean& lhs, const DenseUPolyClean& g, const DenseUPolyClean& h)  // result is true iff quotient is exact.
  {
    if (IsZero(h)) return false; //???CoCoA_ERROR(ERR::DivByZero,"div(DMP,DMP,DMP)");
    CoCoA_ERROR(ERR::NYI, "sub(DenseUPolyClean&, const DenseUPolyClean&, const DenseUPolyClean&)");
    return true;
  }


  void DenseUPolyClean::myAddMul(ConstRefRingElem c, std::size_t d, const DenseUPolyClean& g)
  { // inefficient
    DenseUPolyClean ans(g);
    ans.myMulByCoeff(c);
    ans.myMulByXExp(d);
    add(*this, *this, ans);
  }


  void DenseUPolyClean::myMulByCoeff(ConstRefRingElem c)
  {
    for (size_t i=0 ; i < myDegPlus1Value ; ++i)
      myCoeffsValue[i] *= c;
    if (!IsIntegralDomain(myCoeffRingValue))
      myResetDeg();
  }


  void DenseUPolyClean::myMulByXExp(int n)
  { // EXCEPTION SAFE
    myResize(myDegPlus1Value + n);
    for (size_t i=myDegPlus1Value ; i<mySizeValue  ; ++i)
      myCoeffsValue[i] = 0;
    for (size_t i=myDegPlus1Value; i!=0  ; )
    {
      --i;
      swap(myCoeffsValue[i+n], myCoeffsValue[i]);
    }
    myDegPlus1Value = myDegPlus1Value + n;
  }


  void output(ostream& out, const DenseUPolyClean& f)  // for debugging only
  {
    if (IsZero(f)) { out << "0"; return; }
    for (size_t i=0; i<f.myDegPlus1Value; ++i)
      out << " +(" << f.myCoeffsValue[i] << ")*indet^" << i;
  }


  bool IsZero(const DenseUPolyClean& f)
  {
    return f.myDegPlus1Value == 0;
  }


  bool IsMonomial(const DenseUPolyClean& f)
  {
    CoCoA_ERROR(ERR::NYI, "IsMonomial(const DenseUPolyClean&)");
    return true;
  }


  bool IsEqual(const DenseUPolyClean& f, const DenseUPolyClean& g)
  {
    CoCoA_ASSERT(IsCompatible(f, g));
    if (&f == &g) return true;
    if (f.myDegPlus1Value != g.myDegPlus1Value) return false;
    for (size_t i=0; i < f.myDegPlus1Value; ++i)
      if (f.myCoeffsValue[i] != g.myCoeffsValue[i]) return false;
    return true;
  }


  void deriv(DenseUPolyClean& lhs, const DenseUPolyClean& f)
  {
    CoCoA_ERROR(ERR::NYI, "deriv(DenseUPolyClean&, const DenseUPolyClean&)");
  }

} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/DenseUPolyClean.C,v 1.1 2007/10/05 15:28:56 bigatti Exp $
// $Log: DenseUPolyClean.C,v $
// Revision 1.1  2007/10/05 15:28:56  bigatti
// -- added abstract class DenseUPolyRing for representing dense
//    univariate polynomials
// -- added concrete class RingDenseUPolyClean, the cleanest
//    implementation
// -- just for testing, still horribly inefficient and incomplete
//
